@@ -5,6 +5,7 @@ A trainer class to handle training and testing of models.
 import sys
 import numpy as np
 from collections import Counter
+import logging
 import torch
 from torch import nn
 import torch.nn.init as init
@@ -15,6 +16,8 @@ from classla.models.common import utils, loss
 from classla.models.lemma import edit
 from classla.models.lemma.vocab import MultiVocab
 
+logger = logging.getLogger('classla')
+
 def unpack_batch(batch, use_cuda):
     """ Unpack a batch from the data loader. """
     if use_cuda:
@@ -37,12 +40,12 @@ class Trainer(object):
             self.model = None if args['dict_only'] else Seq2SeqModel(args, emb_matrix=emb_matrix, use_cuda=use_cuda)
             self.vocab = vocab
             # dict-based components
-            self.fallback_dict = dict()
+            self.word_dict = dict()
             self.composite_dict = dict()
         if not self.args['dict_only']:
             if self.args.get('edit', False):
                 self.crit = loss.MixLoss(self.vocab['char'].size, self.args['alpha'])
-                print("[Running seq2seq lemmatizer with edit classifier]")
+                logger.debug("Running seq2seq lemmatizer with edit classifier...")
             else:
                 self.crit = loss.SequenceLoss(self.vocab['char'].size)
             self.parameters = [p for p in self.model.parameters() if p.requires_grad]
@@ -57,6 +60,7 @@ class Trainer(object):
     def update(self, batch, eval=False):
         inputs, orig_idx = unpack_batch(batch, self.use_cuda)
         src, src_mask, tgt_in, tgt_out, pos, edits = inputs
+
         if eval:
             self.model.eval()
         else:
@@ -122,7 +126,7 @@ class Trainer(object):
         utils.change_lr(self.optimizer, new_lr)
 
     def train_dict(self, triples):
-        """ Train a dict lemmatizer given training (word, pos, lemma) tuples. """
+        """ Train a dict lemmatizer given training (word, pos, lemma) triples. """
         # accumulate counter
         ctr = Counter()
         ctr.update([(p[0], p[1], p[2]) for p in triples])
@@ -131,8 +135,8 @@ class Trainer(object):
             w, pos, l = p
             if (w,pos) not in self.composite_dict:
                 self.composite_dict[(w,pos)] = l
-            #if (w, pos) not in self.fallback_dict:
-            #    self.fallback_dict[(w,pos)] = l
+            if w not in self.word_dict:
+                self.word_dict[w] = l
         return
 
     def predict_dict(self, pairs):
@@ -142,8 +146,8 @@ class Trainer(object):
             w, pos = p
             if (w,pos) in self.composite_dict:
                 lemmas += [self.composite_dict[(w,pos)]]
-            #elif (w,pos) in self.fallback_dict:
-            #    lemmas += [self.fallback_dict[(w,pos)]]
+            elif w in self.word_dict:
+                lemmas += [self.word_dict[w]]
             else:
                 lemmas += [w]
         return lemmas
@@ -156,51 +160,65 @@ class Trainer(object):
             w, pos = p
             if (w,pos) in self.composite_dict:
                 skip.append(True)
-            #elif w in self.fallback_dict:
-            #    skip.append(True)
+            elif w in self.word_dict:
+                skip.append(True)
             else:
                 skip.append(False)
         return skip
 
     def ensemble(self, pairs, other_preds):
-        """ Ensemble the dict with statitical model predictions. """
+        """ Ensemble the dict with statistical model predictions. """
         lemmas = []
         assert len(pairs) == len(other_preds)
         for p, pred in zip(pairs, other_preds):
             w, pos = p
             if (w,pos) in self.composite_dict:
-                lemmas += [self.composite_dict[(w,pos)]]
-            #elif (w,pos) in self.fallback_dict:
-            #    lemmas += [self.fallback_dict[(w,pos)]]
+                lemma = self.composite_dict[(w,pos)]
+            elif w in self.word_dict:
+                lemma = self.word_dict[w]
+            elif any([c not in self.vocab['char'] for c in w]):
+                lemma = w
             else:
-                lemmas += [pred]
+                lemma = pred
+            if lemma is None:
+                lemma = w
+            lemmas.append(lemma)
         return lemmas
 
     def save(self, filename):
         params = {
                 'model': self.model.state_dict() if self.model is not None else None,
-                'dicts': (self.fallback_dict, self.composite_dict),
+                'dicts': (self.word_dict, self.composite_dict),
                 'vocab': self.vocab.state_dict(),
                 'config': self.args
                 }
         try:
             torch.save(params, filename)
-            print("model saved to {}".format(filename))
+            logger.info("Model saved to {}".format(filename))
         except BaseException:
-            print("[Warning: Saving failed... continuing anyway.]")
+            logger.warning("Saving failed... continuing anyway.")
 
     def load(self, filename, use_cuda=False):
         try:
             checkpoint = torch.load(filename, lambda storage, loc: storage)
         except BaseException:
-            print("Cannot load model from {}".format(filename))
-            sys.exit(1)
+            logger.error("Cannot load model from {}".format(filename))
+            raise
         self.args = checkpoint['config']
-        self.fallback_dict, self.composite_dict = checkpoint['dicts']
+        self.word_dict, self.composite_dict = checkpoint['dicts']
+
+        # temporary workaround for Bulgarian
+        if ('.', 'Z') not in self.composite_dict:
+            new_elements = {}
+            for k, v in self.composite_dict.items():
+                if k[1] == 'punct':
+                    new_elements[(k[0], 'Z')] = v
+            for k, v in new_elements.items():
+                self.composite_dict[k] = v
+
         if not self.args['dict_only']:
             self.model = Seq2SeqModel(self.args, use_cuda=use_cuda)
             self.model.load_state_dict(checkpoint['model'])
         else:
             self.model = None
         self.vocab = MultiVocab.load_state_dict(checkpoint['vocab'])
-
