@@ -23,6 +23,9 @@ from classla.models.lemma.trainer import Trainer
 from classla.models.lemma import scorer, edit
 from classla.models.common import utils
 import classla.models.common.seq2seq_constant as constant
+from classla.models.common.doc import *
+from classla.utils.conll import CoNLL
+from classla.models import _training_logging
 
 def parse_args():
     parser = argparse.ArgumentParser()
@@ -54,6 +57,7 @@ def parse_args():
     parser.add_argument('--no_edit', dest='edit', action='store_false', help='Do not use edit classifier in lemmatization. By default use an edit classifier.')
     parser.add_argument('--num_edit', type=int, default=len(edit.EDIT_TO_ID))
     parser.add_argument('--alpha', type=float, default=1.0)
+    parser.add_argument('--no_pos', dest='pos', action='store_false', help='Do not use UPOS in lemmatization. By default UPOS is used.')
 
     parser.add_argument('--sample_train', type=float, default=1.0, help='Subsample training data.')
     parser.add_argument('--optim', type=str, default='adam', help='sgd, adagrad, adam or adamax.')
@@ -89,10 +93,6 @@ def main():
     args = vars(args)
     print("Running lemmatizer in {} mode".format(args['mode']))
 
-    # manually correct for training epochs
-    if args['lang'] in ['cs_pdt', 'ru_syntagrus']:
-        args['num_epoch'] = 30
-
     if args['mode'] == 'train':
         train(args)
     else:
@@ -101,14 +101,18 @@ def main():
 def train(args):
     # load data
     print("[Loading data with batch size {}...]".format(args['batch_size']))
-    train_batch = DataLoader(args['train_file'], args['batch_size'], args, evaluation=False)
+    doc, metasentences = CoNLL.conll2dict(input_file=args['train_file'])
+    train_doc = Document(doc, metasentences=metasentences)
+    train_batch = DataLoader(train_doc, args['batch_size'], args, evaluation=False)
     vocab = train_batch.vocab
     args['vocab_size'] = vocab['char'].size
     args['pos_vocab_size'] = vocab['pos'].size
-    dev_batch = DataLoader(args['eval_file'], args['batch_size'], args, vocab=vocab, evaluation=True)
+    doc, metasentences = CoNLL.conll2dict(input_file=args['eval_file'])
+    dev_doc = Document(doc, metasentences=metasentences)
+    dev_batch = DataLoader(dev_doc, args['batch_size'], args, vocab=vocab, evaluation=True)
 
     utils.ensure_dir(args['model_dir'])
-    model_file = '{}/{}_lemmatizer.pt'.format(args['model_dir'], args['model_file'])
+    model_file = '{}/{}_lemmatizer.pt'.format(args['model_dir'], args['lang'])
 
     # pred and gold path
     system_pred_file = args['output_file']
@@ -125,18 +129,19 @@ def train(args):
     # train a dictionary-based lemmatizer
     trainer = Trainer(args=args, vocab=vocab, use_cuda=args['cuda'])
     print("[Training dictionary-based lemmatizer...]")
-    dict = train_batch.conll.get(['word', 'xpos', 'lemma'])
+    dict = train_batch.doc.get([TEXT, XPOS, LEMMA])
     dict = [(e[0].lower(), e[1], e[2]) for e in dict]
     if args.get('external_dict', None) is not None:
         extra_dict = []
         for line in open(args['external_dict']):
-            word,lemma,xpos = line.rstrip('\r\n').split('\t')
-            extra_dict.append((word.lower(),xpos,lemma))
-        dict = extra_dict + dict
+            word, lemma, xpos = line.rstrip('\r\n').split('\t')
+            extra_dict.append((word.lower(), xpos, lemma))
+            dict = extra_dict + dict
     trainer.train_dict(dict)
     print("Evaluating on dev set...")
-    dev_preds = trainer.predict_dict([(e[0].lower(),e[1]) for e in dev_batch.conll.get(['word', 'xpos'])])
-    dev_batch.conll.write_conll_with_lemmas(dev_preds, system_pred_file)
+    dev_preds = trainer.predict_dict([(e[0].lower(), e[1]) for e in dev_batch.doc.get([TEXT, XPOS])])
+    dev_batch.doc.set([LEMMA], dev_preds)
+    CoNLL.dict2conll(dev_batch.doc.to_dict(), system_pred_file)
     _, _, dev_f = scorer.score(system_pred_file, gold_file)
     print("Dev F1 = {:.2f}".format(dev_f * 100))
 
@@ -172,20 +177,27 @@ def train(args):
             dev_preds = []
             dev_edits = []
 
-            # try speeding up dev eval
-            dict_preds = trainer.predict_dict([(e[0].lower(),e[1]) for e in dev_batch.conll.get(['word', 'xpos'])])
+            dict_preds = trainer.predict_dict([(e[0].lower(), e[1]) for e in dev_batch.doc.get([TEXT, XPOS])])
+            # for i, batch in enumerate(dev_batch):
+            #     preds, edits = trainer.predict(batch, args['beam_size'])
+            #     dev_preds += preds
+            #     if edits is not None:
+            #         dev_edits += edits
+            # dev_preds = trainer.postprocess(dev_batch.doc.get([TEXT]), dev_preds, edits=dev_edits)
+
+            # try ensembling with dict if necessary
             if args.get('ensemble_dict', False):
-                skip = trainer.skip_seq2seq([(e[0].lower(),e[1]) for e in dev_batch.conll.get(['word', 'xpos'])])
-                seq2seq_batch = DataLoader(args['eval_file'], args['batch_size'], args, vocab=vocab, evaluation=True, skip=skip)
+                skip = trainer.skip_seq2seq([(e[0].lower(), e[1]) for e in dev_batch.doc.get([TEXT, XPOS])])
+                doc, metasentences = CoNLL.conll2dict(input_file=args['eval_file'])
+                dev_doc = Document(doc, metasentences=metasentences)
+                seq2seq_batch = DataLoader(dev_doc, args['batch_size'], args, vocab=vocab, evaluation=True, skip=skip)
+                # print("[Ensembling dict with seq2seq model...]")
+                # dev_preds = trainer.ensemble(dev_batch.doc.get([TEXT, UPOS]), dev_preds)
             else:
                 seq2seq_batch = dev_batch
-            for i, b in enumerate(seq2seq_batch):
-                ps, es = trainer.predict(b, args['beam_size'])
-                dev_preds += ps
-                if es is not None:
-                    dev_edits += es
+
             if args.get('ensemble_dict', False):
-                dev_preds = trainer.postprocess([x for x, y in zip(dev_batch.conll.get(['word']), skip) if not y], dev_preds, edits=dev_edits)
+                dev_preds = trainer.postprocess([x for x, y in zip(dev_batch.doc.get([TEXT]), skip) if not y], dev_preds, edits=dev_edits)
                 print("[Ensembling dict with seq2seq lemmatizer...]")
                 i = 0
                 preds1 = []
@@ -195,12 +207,13 @@ def train(args):
                     else:
                         preds1.append(dev_preds[i])
                         i += 1
-                dev_preds = trainer.ensemble([(e[0].lower(),e[1]) for e in dev_batch.conll.get(['word', 'xpos'])], preds1)
+                dev_preds = trainer.ensemble([(e[0].lower(),e[1]) for e in dev_batch.doc.get([TEXT, XPOS])], preds1)
             else:
-                dev_preds = trainer.postprocess(dev_batch.conll.get(['word']), dev_preds, edits=dev_edits)
+                dev_preds = trainer.postprocess(dev_batch.doc.get([TEXT]), dev_preds, edits=dev_edits)
 
 
-            dev_batch.conll.write_conll_with_lemmas(dev_preds, system_pred_file)
+            dev_batch.doc.set([LEMMA], dev_preds)
+            CoNLL.dict2conll(dev_batch.doc.to_dict(), system_pred_file)
             _, _, dev_score = scorer.score(system_pred_file, gold_file)
 
             train_loss = train_loss / train_batch.num_examples * args['batch_size'] # avg loss per batch
@@ -230,7 +243,7 @@ def evaluate(args):
     # file paths
     system_pred_file = args['output_file']
     gold_file = args['gold_file']
-    model_file = '{}/{}_lemmatizer.pt'.format(args['model_dir'], args['model_file'])
+    model_file = '{}/{}.pt'.format(args['model_dir'], args['model_file'])
 
     # load model
     use_cuda = args['cuda'] and not args['cpu']
@@ -243,7 +256,9 @@ def evaluate(args):
 
     # laod data
     print("Loading data with batch size {}...".format(args['batch_size']))
-    batch = DataLoader(args['eval_file'], args['batch_size'], loaded_args, vocab=vocab, evaluation=True)
+    doc, metasentences = CoNLL.conll2dict(input_file=args['eval_file'])
+    doc = Document(doc, metasentences=metasentences)
+    batch = DataLoader(doc, args['batch_size'], loaded_args, vocab=vocab, evaluation=True)
 
     # skip eval if dev data does not exist
     if len(batch) == 0:
@@ -252,14 +267,16 @@ def evaluate(args):
         print("{} ".format(args['lang']))
         sys.exit(0)
 
-    dict_preds = trainer.predict_dict([(e[0].lower(),e[1]) for e in batch.conll.get(['word', 'xpos'])])
+    dict_preds = trainer.predict_dict([(e[0].lower(),e[1]) for e in batch.doc.get([TEXT, XPOS])])
 
     if loaded_args.get('dict_only', False):
         preds = dict_preds
     else:
         if loaded_args.get('ensemble_dict', False):
-            skip = trainer.skip_seq2seq([(e[0].lower(),e[1]) for e in batch.conll.get(['word', 'xpos'])])
-            seq2seq_batch = DataLoader(args['eval_file'], args['batch_size'], loaded_args, vocab=vocab, evaluation=True, skip=skip)
+            skip = trainer.skip_seq2seq([(e[0].lower(),e[1]) for e in batch.doc.get([TEXT, XPOS])])
+            doc, metasentences = CoNLL.conll2dict(input_file=args['eval_file'])
+            dev_doc = Document(doc, metasentences=metasentences)
+            seq2seq_batch = DataLoader(dev_doc, args['batch_size'], loaded_args, vocab=vocab, evaluation=True, skip=skip)
         else:
             seq2seq_batch = batch
         print("Running the seq2seq model...")
@@ -270,9 +287,11 @@ def evaluate(args):
             preds += ps
             if es is not None:
                 edits += es
+        # preds = trainer.postprocess(batch.doc.get([TEXT]), preds, edits=edits)
 
         if loaded_args.get('ensemble_dict', False):
-            preds = trainer.postprocess([x for x, y in zip(batch.conll.get(['word']), skip) if not y], preds, edits=edits)
+            preds = trainer.postprocess([x for x, y in zip(batch.doc.get([TEXT]), skip) if not y], preds,
+                                        edits=edits)
             print("[Ensembling dict with seq2seq lemmatizer...]")
             i = 0
             preds1 = []
@@ -282,12 +301,15 @@ def evaluate(args):
                 else:
                     preds1.append(preds[i])
                     i += 1
-            preds = trainer.ensemble([(e[0].lower(),e[1]) for e in batch.conll.get(['word', 'xpos'])], preds1)
+            preds = trainer.ensemble([(e[0].lower(), e[1]) for e in batch.doc.get([TEXT, XPOS])], preds1)
         else:
-            preds = trainer.postprocess(batch.conll.get(['word']), preds, edits=edits)
+            preds = trainer.postprocess(batch.doc.get([TEXT]), preds, edits=edits)
+            print("[Ensembling dict with seq2seq lemmatizer...]")
+            # preds = trainer.ensemble(batch.doc.get([TEXT, UPOS]), preds)
 
     # write to file and score
-    batch.conll.write_conll_with_lemmas(preds, system_pred_file)
+    batch.doc.set([LEMMA], preds)
+    CoNLL.dict2conll(batch.doc.to_dict(), system_pred_file)
     if gold_file is not None:
         _, _, score = scorer.score(system_pred_file, gold_file)
 
