@@ -13,6 +13,7 @@ from classla.models.ner.model import NERTagger
 from classla.models.ner.vocab import MultiVocab
 from classla.models.common.crf import viterbi_decode
 
+logger = logging.getLogger('classla')
 
 def unpack_batch(batch, use_cuda):
     """ Unpack a batch from the data loader. """
@@ -104,18 +105,18 @@ class Trainer(BaseTrainer):
                 }
         try:
             torch.save(params, filename)
-            print("Model saved to {}".format(filename))
+            logger.info("Model saved to {}".format(filename))
         except (KeyboardInterrupt, SystemExit):
             raise
         except:
-            print("Saving failed... continuing anyway.")
+            logger.warning("Saving failed... continuing anyway.")
 
     def load(self, filename, args=None):
         try:
             checkpoint = torch.load(filename, lambda storage, loc: storage)
         except BaseException:
-            print("Cannot load model from {}".format(filename))
-            sys.exit(1)
+            logger.error("Cannot load model from {}".format(filename))
+            raise
         self.args = checkpoint['config']
         if args: self.args.update(args)
         self.vocab = MultiVocab.load_state_dict(checkpoint['vocab'])
