@@ -4,7 +4,13 @@ import re
 from classla.models.common.vocab import BaseVocab
 from classla.models.common.vocab import UNK, PAD
 
+SPACE_RE = re.compile(r'\s')
+
 class Vocab(BaseVocab):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.lang_replaces_spaces = any([self.lang.startswith(x) for x in ['zh', 'ja', 'ko']])
+
     def build_vocab(self):
         paras = self.data
         counter = Counter()
@@ -26,9 +32,9 @@ class Vocab(BaseVocab):
         return normalized
 
     def normalize_token(self, token):
-        token = re.sub('\s', ' ', token.lstrip())
+        token = SPACE_RE.sub(' ', token.lstrip())
 
-        if any([self.lang.startswith(x) for x in ['zh', 'ja', 'ko']]):
+        if self.lang_replaces_spaces:
             token = token.replace(' ', '')
 
         return token
