@@ -2,18 +2,19 @@
 Processor for performing lemmatization
 """
 
-from classla.models.common.conll import FIELD_TO_IDX
+from classla.models.common import doc
 from classla.models.lemma.data import DataLoader
 from classla.models.lemma.trainer import Trainer
 from classla.pipeline._constants import *
-from classla.pipeline.processor import UDProcessor
-
+from classla.pipeline.processor import UDProcessor, register_processor
 
+@register_processor(name=LEMMA)
 class LemmaProcessor(UDProcessor):
 
     # set of processor requirements this processor fulfills
     PROVIDES_DEFAULT = set([LEMMA])
     # set of processor requirements for this processor
+    # pos will be added later for non-identity lemmatizerx
     REQUIRES_DEFAULT = set([TOKENIZE])
     # default batch size
     DEFAULT_BATCH_SIZE = 5000
@@ -42,20 +43,20 @@ class LemmaProcessor(UDProcessor):
         else:
             self._requires = LemmaProcessor.REQUIRES_DEFAULT
 
-    def process(self, doc):
+    def process(self, document):
         if not self.use_identity:
-            batch = DataLoader(doc, self.config['batch_size'], self.config, vocab=self.vocab, evaluation=True)
+            batch = DataLoader(document, self.config['batch_size'], self.config, vocab=self.vocab, evaluation=True)
         else:
-            batch = DataLoader(doc, self.config['batch_size'], self.config, evaluation=True, conll_only=True)
+            batch = DataLoader(document, self.config['batch_size'], self.config, evaluation=True, conll_only=True)
         if self.use_identity:
-            preds = [ln[FIELD_TO_IDX['word']] for sent in batch.conll.sents for ln in sent if '-' not in ln[0]]
+            preds = [word.text for sent in batch.doc.sentences for word in sent.words]
         elif self.config.get('dict_only', False):
-            preds = self.trainer.predict_dict(batch.conll.get(['word', 'xpos']))
+            preds = self.trainer.predict_dict(batch.doc.get([doc.TEXT, doc.XPOS]))
         else:
             if self.config.get('ensemble_dict', False):
                 # skip the seq2seq model when we can
-                skip = self.trainer.skip_seq2seq([(e[0].lower(),e[1]) for e in batch.conll.get(['word', 'xpos'])])
-                seq2seq_batch = DataLoader(doc, self.config['batch_size'], self.config, vocab=self.vocab,
+                skip = self.trainer.skip_seq2seq([(e[0].lower(),e[1]) for e in batch.doc.get([doc.TEXT, doc.XPOS])])
+                seq2seq_batch = DataLoader(document, self.config['batch_size'], self.config, vocab=self.vocab,
                                            evaluation=True, skip=skip)
             else:
                 seq2seq_batch = batch
@@ -69,7 +70,7 @@ class LemmaProcessor(UDProcessor):
                     edits += es
 
             if self.config.get('ensemble_dict', False):
-                preds = self.trainer.postprocess([x for x, y in zip(batch.conll.get(['word']), skip) if not y], preds, edits=edits)
+                preds = self.trainer.postprocess([x for x, y in zip(batch.doc.get([doc.TEXT]), skip) if not y], preds, edits=edits)
                 # expand seq2seq predictions to the same size as all words
                 i = 0
                 preds1 = []
@@ -79,11 +80,11 @@ class LemmaProcessor(UDProcessor):
                     else:
                         preds1.append(preds[i])
                         i += 1
-                preds = self.trainer.ensemble([(e[0].lower(),e[1]) for e in batch.conll.get(['word', 'xpos'])], preds1)
+                preds = self.trainer.ensemble([(e[0].lower(),e[1]) for e in batch.doc.get([doc.TEXT, doc.XPOS])], preds1)
             else:
-                preds = self.trainer.postprocess(batch.conll.get(['word']), preds, edits=edits)
+                preds = self.trainer.postprocess(batch.doc.get([doc.TEXT]), preds, edits=edits)
 
         # map empty string lemmas to '_'
         preds = [max([(len(x), x), (0, '_')])[1] for x in preds]
-        batch.conll.set(['lemma'], preds)
-
+        batch.doc.set([doc.LEMMA], preds)
+        return batch.doc
