@@ -6,5 +6,5 @@ from classla.protobuf import Quote, SpeakerInfo
 from classla.protobuf import Operator, Polarity
 from classla.protobuf import SentenceFragment, TokenLocation
 from classla.protobuf import MapStringString, MapIntString
-from .client import CoreNLPClient, AnnotationException, TimeoutException
+from .client import CoreNLPClient, AnnotationException, TimeoutException, PermanentlyFailedException, StartServer
 from .annotator import Annotator
